# Copyright (C) 2011-2014 by Barry A. Warsaw
#
# This file is part of flufl.password
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

"""Password hashing and verification schemes.

Represents passwords using RFC 2307 syntax (as best we can tell).
"""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'ClearTextPasswordScheme',
    'NoPasswordScheme',
    'PBKDF2PasswordScheme',
    'PasswordScheme',
    'SHAPasswordScheme',
    'SSHAPasswordScheme',
    ]


import os
import hmac
import hashlib

from array import array
from base64 import urlsafe_b64decode as decode
from base64 import urlsafe_b64encode as encode
from functools import partial

from flufl.password._registry import register
from flufl.password._utils import BadPasswordFormatError


SALT_LENGTH = 20 # bytes
ITERATIONS  = 2000



class PasswordScheme:
    """Password scheme base class."""
    TAG = ''

    @staticmethod
    def make_secret(password):
        """Return the hashed password.

        :param password: The clear text password.
        :type password: bytes or utf-8 encoded string.
        :return: The encrypted password.
        :rtype: bytes
        """
        raise NotImplementedError

    @classmethod
    def check_response(cls, hashed, cleartext, scheme=None):
        """Check a hashed password against a clear text response.

        :param hashed: The user's stored hashed password, without the scheme
            prefix (i.e. strip off the *{SCHEME}* part first).
        :type hashed: bytes
        :param cleartext: The clear text password entered by the user for
            authentication.
        :type cleartext: bytes or utf-8 encoded string.
        :param scheme: The scheme tag i.e. the *{SCHEME}* part without the
            braces.  This can be omitted when called on the class directly,
            unless the scheme puts some additional information in there.  It
            is always provided by the `verify()` API.
        :type response: bytes or utf-8 encoded string.
        :return: True if the clear text matches the hash
        :rtype: bool
        """
        return hashed == cls.make_secret(cleartext)



@register
class NoPasswordScheme(PasswordScheme):
    """A password scheme without passwords."""

    TAG = 'NONE'

    @staticmethod
    def make_secret(password):
        """See `PasswordScheme`."""
        return b''

    @classmethod
    def check_response(cls, hashed, cleartext, scheme=None):
        """See `PasswordScheme`."""
        return False



@register
class ClearTextPasswordScheme(PasswordScheme):
    """A password scheme that stores clear text passwords."""

    TAG = 'CLEARTEXT'

    @staticmethod
    def make_secret(password):
        """See `PasswordScheme`."""
        return password



@register
class SHAPasswordScheme(PasswordScheme):
    """A password scheme that encodes the password using SHA1."""

    TAG = 'SHA'

    @staticmethod
    def make_secret(password):
        """See `PasswordScheme`."""
        h = hashlib.sha1(password)
        return encode(h.digest())



@register
class SSHAPasswordScheme(PasswordScheme):
    """A password scheme that encodes the password using salted SHA1."""

    TAG = 'SSHA'

    @staticmethod
    def _make_salted_hash(password, salt):
        h = hashlib.sha1(password)
        h.update(salt)
        return encode(h.digest() + salt)

    @staticmethod
    def make_secret(password):
        """See `PasswordScheme`."""
        salt = os.urandom(SALT_LENGTH)
        return SSHAPasswordScheme._make_salted_hash(password, salt)

    @classmethod
    def check_response(cls, hashed, cleartext, scheme=None):
        """See `PasswordScheme`."""
        # The salt is always 20 bytes and always tacked onto the end.
        decoded = decode(hashed)
        salt = decoded[20:]
        return hashed == cls._make_salted_hash(cleartext, salt)



def _bytes_of(array_obj):
    # Avoid DeprecationWarnings in Python 3.
    try:
        return array_obj.tobytes()
    except AttributeError:
        return array_obj.tostring()


# Basic algorithm given by Bob Fleck
@register
class PBKDF2PasswordScheme(PasswordScheme):
    """RFC 2989 password encoding scheme."""

    # This is a bit nasty if we wanted a different prf or iterations.  OTOH,
    # we really have no clue what the standard LDAP-ish specification for
    # those options is.
    TAG = 'PBKDF2 SHA {0}'.format(ITERATIONS)

    @staticmethod
    def _pbkdf2(password, salt, iterations):
        """From RFC2898 sec. 5.2.  Simplified to handle only 20 byte output
        case.  Output of 20 bytes means always exactly one block to handle,
        and a constant block counter appended to the salt in the initial hmac
        update.
        """
        # We do it this way because the array() constructor takes a 'native
        # string'.  You can't use unicodes in Python 2 or bytes in Python 3.
        array_of_signedints = partial(array, str('i'))
        h = hmac.new(password, None, hashlib.sha1)
        prf = h.copy()
        prf.update(salt + b'\x00\x00\x00\x01')
        T = U = array_of_signedints(prf.digest())
        while iterations:
            prf = h.copy()
            prf.update(_bytes_of(U))
            U = array_of_signedints(prf.digest())
            T = array_of_signedints((t ^ u for t, u in zip(T, U)))
            iterations -= 1
        return _bytes_of(T)

    @staticmethod
    def make_secret(password):
        """See `PasswordScheme`.

        From RFC2898 sec. 5.2.  Simplified to handle only 20 byte output
        case.  Output of 20 bytes means always exactly one block to handle,
        and a constant block counter appended to the salt in the initial hmac
        update.
        """
        salt = os.urandom(SALT_LENGTH)
        digest = PBKDF2PasswordScheme._pbkdf2(password, salt, ITERATIONS)
        return encode(digest + salt)

    @classmethod
    def check_response(cls, hashed, cleartext, scheme):
        """See `PasswordScheme`."""
        parts = scheme.split()
        if (len(parts) != 3 or
            parts[0].upper() != 'PBKDF2' or
            parts[1].upper() != 'SHA'
            ):
            raise BadPasswordFormatError(scheme)
        try:
            iterations = int(parts[2])
        except ValueError:
            raise BadPasswordFormatError(scheme)
        decoded = decode(hashed)
        salt = decoded[20:]
        secret = decoded[:20]
        return secret == cls._pbkdf2(cleartext, salt, iterations)



# Copyright (C) 2011-2014 by Barry A. Warsaw
#
# This file is part of flufl.password
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

"""Scheme registry."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'BadPasswordSchemeError',
    'lookup',
    'register',
    ]


_registry = {}



class BadPasswordSchemeError(Exception):
    """An unknown password scheme tag was given."""



def register(scheme_class):
    """Register a scheme by its tag.

    This is intended to be used as a class descriptor.

    :param scheme_class: A password scheme.
    :type scheme_class: A class which has a `TAG` attribute.  The tag may
        contain space-separated words with additional information, but the
        scheme will be registered using the first word as the key.
    :return: scheme_class
    """
    key = scheme_class.TAG.split()[0]
    _registry[key] = scheme_class
    return scheme_class


def lookup(tag):
    """Return a scheme class by its tag.

    :param tag: The tag to search for.  The tag may contain space-separated
        words with additional information, but the scheme will be looked up
        using the first word as the key.
    :type tag: string
    :return: The matching scheme class.
    :rtype: class
    :raises BadPasswordSchemeError: when the named scheme can't be found.
    """
    key = tag.split()[0]
    try:
        return _registry[key]
    except KeyError:
        raise BadPasswordSchemeError(tag)



# Copyright (C) 2011-2014 by Barry A. Warsaw
#
# This file is part of flufl.password
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

"""User-friendly password generation."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'generate',
    ]


import random

from itertools import chain, product
from string import ascii_lowercase


EMPTYSTRING = ''



_vowels = tuple('aeiou')
_consonants = tuple(c for c in ascii_lowercase if c not in _vowels)
_syllables = tuple(x + y for (x, y) in
                   chain(product(_vowels, _consonants),
                         product(_consonants, _vowels)))



def generate(length=8):
    """Make a random *user friendly* password.

    Such passwords are nominally easier to pronounce and thus remember.  Their
    security in relationship to purely random passwords has not been
    determined.

    :param length: Minimum length in characters for the resulting password.
        The password will always be an even number of characters.  The default
        is to create passwords of length 8.
    :type length: int
    :return: The user friendly password.
    :rtype: string
    """
    syllables = []
    while len(syllables) * 2 < length:
        syllables.append(random.choice(_syllables))
    return EMPTYSTRING.join(syllables)[:length]



# Copyright (C) 2014 by Barry A. Warsaw
#
# This file is part of flufl.password
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

"""Module contents."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    ]



# Copyright (C) 2011-2014 by Barry A. Warsaw
#
# This file is part of flufl.password
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for the passwords module."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'TestCleartextPasswords',
    'TestNoPasswords',
    'TestPBKDF2Passwords',
    'TestPasswordGeneration',
    'TestSHAPasswords',
    'TestSSHAPasswords',
    'TestSchemeLookup',
    ]


import unittest

# Python 3 does not have the izip form.
try:
    from itertools import izip_longest as longest
except ImportError:
    from itertools import zip_longest as longest

from flufl.password._hash import make_secret
from flufl.password._registry import BadPasswordSchemeError, lookup
from flufl.password._verify import verify
from flufl.password import schemes, generate



class PasswordsTestBase:
    scheme = None

    def setUp(self):
        # The user's password, as a bytes
        self.pwbyte     = b'abc'
        self.pwutf8     = 'abc\xc3\xbf'             # 'abc\xff'
        # Bad passwords; bytes
        self.badbyte    = b'def'

    def test_byte_passwords(self):
        secret = make_secret(self.pwbyte, self.scheme)
        self.assertTrue(verify(secret, self.pwbyte), self.scheme)
        self.assertFalse(verify(secret, self.badbyte), self.scheme)

    def test_utf8_passwords(self):
        secret = make_secret(self.pwutf8, self.scheme)
        self.assertTrue(verify(secret, self.pwutf8), self.scheme)
        self.assertFalse(verify(secret, self.badbyte), self.scheme)



class TestNoPasswords(unittest.TestCase):
    def test_make_secret(self):
        self.assertEqual(schemes.NoPasswordScheme.make_secret('whatever'), b'')

    def test_check_response(self):
        self.assertFalse(
            schemes.NoPasswordScheme.check_response(b'foo', 'bar'))
        self.assertFalse(schemes.NoPasswordScheme.check_response(b'', 'bar'))



class TestCleartextPasswords(PasswordsTestBase, unittest.TestCase):
    scheme = schemes.ClearTextPasswordScheme


class TestSHAPasswords(PasswordsTestBase, unittest.TestCase):
    scheme = schemes.SHAPasswordScheme


class TestSSHAPasswords(PasswordsTestBase, unittest.TestCase):
    scheme = schemes.SSHAPasswordScheme


class TestPBKDF2Passwords(PasswordsTestBase, unittest.TestCase):
    scheme = schemes.PBKDF2PasswordScheme



class TestSchemeLookup(unittest.TestCase):
    def test_scheme_name_lookup(self):
        self.assertEqual(lookup('NONE'), schemes.NoPasswordScheme)
        self.assertEqual(lookup('CLEARTEXT'), schemes.ClearTextPasswordScheme)
        self.assertEqual(lookup('SHA'), schemes.SHAPasswordScheme)
        self.assertEqual(lookup('SSHA'), schemes.SSHAPasswordScheme)
        self.assertEqual(lookup('PBKDF2'), schemes.PBKDF2PasswordScheme)

    def test_lookup_error(self):
        self.assertRaises(BadPasswordSchemeError, lookup, 'BOGUS')




# See itertools doc page examples.
def _grouper(seq):
    args = [iter(seq)] * 2
    return list(longest(*args))


class TestPasswordGeneration(unittest.TestCase):
    def test_provided_user_friendly_password_length(self):
        self.assertEqual(len(generate(12)), 12)

    def test_provided_odd_user_friendly_password_length(self):
        self.assertEqual(len(generate(15)), 15)

    def test_user_friendly_password(self):
        password = generate()
        for pair in _grouper(password):
            # There will always be one vowel and one non-vowel.
            vowel = (pair[0] if pair[0] in 'aeiou' else pair[1])
            consonant = (pair[0] if pair[0] not in 'aeiou' else pair[1])
            self.assertTrue(vowel in 'aeiou', vowel)
            self.assertTrue(consonant not in 'aeiou', consonant)



# Copyright (C) 2004-2014 by Barry A. Warsaw
#
# This file is part of flufl.password.
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# flufl.password is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

# this is a namespace package
try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil
    __path__ = pkgutil.extend_path(__path__, __name__)



# Copyright (C) 2004-2014 by Barry A. Warsaw
#
# This file is part of flufl.password
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

"""Package init."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'BadPasswordFormatError',
    'BadPasswordSchemeError',
    '__version__',
    'generate',
    'lookup',
    'make_secret',
    'register',
    'verify',
    ]


__version__ = '1.3'


from ._hash import make_secret
from ._generate import generate
from ._registry import BadPasswordSchemeError, lookup, register
from ._utils import BadPasswordFormatError
from ._verify import verify

# Register the built-in schemes by import, but don't expose this in the API.
# Users should import specific schemes explicitly.
from . import schemes as _schemes



# Copyright (C) 2011-2014 by Barry A. Warsaw
#
# This file is part of flufl.password
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

"""Verification."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'verify',
    ]


from flufl.password._registry import lookup
from flufl.password._utils import split



def verify(hashed, cleartext):
    """Verify whether the clear text response matches the hashed challenge.

    :param hashed: The hashed password.
    :type hashed: bytes
    :param cleartext: The clear text password to match against.
    :type cleartext: bytes or utf-8 encoded string.
    :return: Flag indicating whether the password matched.
    :rtype: bool
    :raises BadPasswordSchemeError: when the indicated scheme cannot be found.
    :raises BadPasswordFormatError: when the hashed parameter is not properly
        formatted as ``{SCHEME}password``.
    """
    if not isinstance(cleartext, bytes):
        cleartext = cleartext.encode('utf-8')
    scheme, secret = split(hashed)
    scheme_class = lookup(scheme.upper())
    return scheme_class.check_response(secret, cleartext, scheme)



# Copyright (C) 2004-2014 by Barry A. Warsaw
#
# This file is part of flufl.password.
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

from setup_helpers import (
    description, get_version, long_description, require_python)
from setuptools import setup, find_packages


require_python(0x20600f0)
__version__ = get_version('flufl/password/__init__.py')


setup(
    name='flufl.password',
    version=__version__,
    namespace_packages=['flufl'],
    packages=find_packages(),
    include_package_data=True,
    maintainer='Barry Warsaw',
    maintainer_email='barry@python.org',
    description=description('README.rst'),
    long_description=long_description('README.rst', 'flufl/password/NEWS.rst'),
    license='LGPLv3',
    url='http://launchpad.net/flufl.password',
    download_url='https://launchpad.net/flufl.password/+download',
    test_suite='flufl.password.tests',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: '
            'GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ]
    )



# Copyright (C) 2004-2014 by Barry A. Warsaw
#
# This file is part of flufl.password.
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

"""Test harness for doctests."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'additional_tests',
    ]


import os
import atexit
import doctest
import unittest

from pkg_resources import (
    resource_filename, resource_exists, resource_listdir, cleanup_resources)


COMMASPACE = ', '
DOT = '.'
DOCTEST_FLAGS = (
    doctest.ELLIPSIS |
    doctest.NORMALIZE_WHITESPACE |
    doctest.REPORT_NDIFF)



def stop():
    """Call into pdb.set_trace()"""
    # Do the import here so that you get the wacky special hacked pdb instead
    # of Python's normal pdb.
    import pdb
    pdb.set_trace()



def show(thing):
    # Python 2 and 3 display bytes objects differently, even with print.  This
    # makes it more difficult to write doctests that work in both versions of
    # Python, because hashed passwords are all bytes.  The root cause is that
    # in Python 2, `bytes` is just an alias for `str` and in Python 3, the
    # repr of bytes includes the b'' prefix.  Use this instead of print() in
    # the doctests.
    if isinstance(thing, bytes) and (bytes is not str):
        # There are lots of ways we could do this.
        thing = repr(thing)[2:-1]
    print(thing)



def setup(testobj):
    """Test setup."""
    # Make sure future statements in our doctests match the Python code.  When
    # run with 2to3, the future import gets removed and these names are not
    # defined.
    try:
        testobj.globs['absolute_import'] = absolute_import
        testobj.globs['print_function'] = print_function
        testobj.globs['unicode_literals'] = unicode_literals
    except NameError:
        pass
    testobj.globs['show'] = show
    testobj.globs['stop'] = stop



def additional_tests():
    "Run the doc tests (README.rst and docs/*, if any exist)"
    doctest_files = [
        os.path.abspath(resource_filename('flufl.password', 'README.rst'))]
    if resource_exists('flufl.password', 'docs'):
        for name in resource_listdir('flufl.password', 'docs'):
            if name.endswith('.rst'):
                doctest_files.append(
                    os.path.abspath(
                        resource_filename('flufl.password', 'docs/%s' % name)))
    kwargs = dict(module_relative=False,
                  optionflags=DOCTEST_FLAGS,
                  setUp=setup,
                  )
    atexit.register(cleanup_resources)
    return unittest.TestSuite((
        doctest.DocFileSuite(*doctest_files, **kwargs)))



# Copyright (C) 2011-2014 by Barry A. Warsaw
#
# This file is part of flufl.password
#
# flufl.password is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# flufl.password is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with flufl.password.  If not, see <http://www.gnu.org/licenses/>.

"""Make secrets."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'make_secret',
    ]



def make_secret(password, scheme=None):
    """Return a hashed password using the given scheme.

    :param password: The plain text password to hash.
    :type password: bytes or utf-8 encoded string.
    :param scheme: The scheme class to use for encryption.  If not given,
        `SSHAPasswordScheme` is used.
    :type scheme: PasswordScheme subclass.
    :return: The hashed password.
    :rtype: bytes
    """
    if not isinstance(password, bytes):
        # If it's not a bytes, it better be a unicode in Python 2 or a str in
        # Python 3.  Let encoding errors propagate up.
        password = password.encode('utf-8')
    if scheme is None:
        from flufl.password.schemes import SSHAPasswordScheme
        scheme = SSHAPasswordScheme
    secret = scheme.make_secret(password)
    if not isinstance(scheme.TAG, bytes):
        tag = scheme.TAG.encode('utf-8')
    return b'{' + tag + b'}' + secret
